import logging
from typing import List, Optional

from _qwak_proto.qwak.offline.serving.v1.feature_values_pb2 import (
    FeaturesetFeatures as ProtoFeaturesetFeatures,
)
from _qwak_proto.qwak.offline.serving.v1.offline_serving_async_service_pb2 import (
    FeatureValuesRequestStatus as ProtoFeatureValuesRequestStatus,
    FileFormat as ProtoFileFormat,
    GetFeatureValuesInRangeRequest as ProtoGetFeatureValuesInRangeRequest,
    GetFeatureValuesInRangeResponse as ProtoGetFeatureValuesInRangeResponse,
    GetFeatureValuesRequest as ProtoGetFeatureValuesRequest,
    GetFeatureValuesResponse as ProtoGetFeatureValuesResponse,
    GetFeatureValuesResultRequest as ProtoGetFeatureValuesResultRequest,
    GetFeatureValuesResultResponse as ProtoGetFeatureValuesResultResponse,
    GetFileUploadUrlRequest as ProtoGetFileUploadUrlRequest,
    GetFileUploadUrlResponse as ProtoGetFileUploadUrlResponse,
)
from _qwak_proto.qwak.offline.serving.v1.offline_serving_async_service_pb2_grpc import (
    FeatureStoreOfflineServingAsyncServiceStub,
)
from _qwak_proto.qwak.offline.serving.v1.options_pb2 import (
    OfflineServingQueryOptions as ProtoOfflineServingQueryOptions,
)
from _qwak_proto.qwak.offline.serving.v1.population_pb2 import (
    Population as ProtoPopulation,
    PopulationFileUploadUrlType as ProtoPopulationFileUploadUrlType,
    TimedPopulation as ProtoTimedPopulation,
)
from dependency_injector.wiring import Provide
from google.protobuf.timestamp_pb2 import Timestamp as ProtoTimestamp
from qwak.inner.di_configuration import QwakContainer
from qwak.inner.tool.retry_utils import retry


class FeatureValuesResultNotReadyException(Exception):
    pass


class FeatureValuesTimeoutException(Exception):
    pass


class FsOfflineServingClient:
    """
    Querying offline features store
    """

    def __init__(self, grpc_channel=Provide[QwakContainer.core_grpc_channel]):
        self._client = FeatureStoreOfflineServingAsyncServiceStub(grpc_channel)

    def get_population_file_upload_url(
        self,
    ) -> str:
        """
        Generating population file upload url
        :return: pre-signed url to upload population data
        """

        request: ProtoGetFileUploadUrlRequest = ProtoGetFileUploadUrlRequest(
            population=ProtoPopulationFileUploadUrlType()
        )

        response: ProtoGetFileUploadUrlResponse = self._client.GetFileUploadUrl(request)
        return response.file_upload_url

    def get_feature_values(
        self,
        features: List[ProtoFeaturesetFeatures],
        population: ProtoTimedPopulation,
        result_file_format: ProtoFileFormat,
        options: ProtoOfflineServingQueryOptions,
    ) -> str:
        """
        Getting offline feature values
        :return: Request handle id to poll for result with
        """
        request: ProtoGetFeatureValuesRequest = ProtoGetFeatureValuesRequest(
            features=features,
            population=population,
            result_file_format=result_file_format,
            options=options,
        )

        response: ProtoGetFeatureValuesResponse = self._client.GetFeatureValues(request)
        return response.request_id

    def get_feature_values_in_range(
        self,
        features: ProtoFeaturesetFeatures,
        lower_time_bound: ProtoTimestamp,
        upper_time_bound: ProtoTimestamp,
        result_file_format: ProtoFileFormat,
        population: Optional[ProtoPopulation],
        options: ProtoOfflineServingQueryOptions,
    ) -> str:
        """
        Getting offline feature values
        :return: Request handle id to poll for result with
        """
        request: ProtoGetFeatureValuesInRangeRequest = (
            ProtoGetFeatureValuesInRangeRequest(
                features=features,
                lower_time_bound=lower_time_bound,
                upper_time_bound=upper_time_bound,
                result_file_format=result_file_format,
                population=population,
                options=options,
            )
        )

        response: ProtoGetFeatureValuesInRangeResponse = (
            self._client.GetFeatureValuesInRange(request)
        )
        return response.request_id

    def get_result(self, request_handle: str) -> ProtoGetFeatureValuesResultResponse:
        request: ProtoGetFeatureValuesResultRequest = (
            ProtoGetFeatureValuesResultRequest(request_id=request_handle)
        )

        return self._client.GetFeatureValuesResult(request)

    def _inner_poll(self, request_handle: str) -> ProtoGetFeatureValuesResultResponse:
        """
        This function receives request_handle amd polls for results.

        Args:
            request_handle: request_id for polling

        Returns: if the request status is PENDING it throws FeatureValuesResultNotReadyException.
                 else returns the Feature Values Result
        """
        response: ProtoGetFeatureValuesResultResponse = self.get_result(
            request_handle=request_handle
        )

        if (
            response.status
            == ProtoFeatureValuesRequestStatus.FEATURE_VALUES_REQUEST_STATUS_PENDING
        ):
            logging.info("Feature Values query is still in progress...")
            raise FeatureValuesResultNotReadyException()

        return response

    def poll_for_result(
        self,
        request_handle: str,
        timeout_seconds: int = 60 * 60,
        poll_interval_seconds: int = 10,
    ) -> ProtoGetFeatureValuesResultResponse:
        """
        This function receives request_handle and polls for Feature Values Result.
        it relies on _inner_poll() to raise FeatureValuesResultNotReadyException in case the result is not ready.
        only once the desired amount of attempts completed it raises FeatureValuesTimeoutException
        Args:
            request_handle: request_id
            timeout_seconds: total number of seconds for all attempts before timeout
            poll_interval_seconds: sleep time between attempts

        Returns: Feature Values Result or raises FeatureValuesTimeoutException if attempts exceeded timeout_seconds.

        """
        try:
            result = retry(
                f=self._inner_poll,
                kwargs={"request_handle": request_handle},
                exceptions=FeatureValuesResultNotReadyException,
                attempts=int(timeout_seconds / poll_interval_seconds) + 1,
                delay=poll_interval_seconds,
            )
        except FeatureValuesResultNotReadyException:
            raise FeatureValuesTimeoutException(
                f"Feature values query timed out. Qwak limits query execution time to {int(timeout_seconds / 60)} minutes"
            )

        return result

    def get_feature_values_blocking(
        self,
        features: List[ProtoFeaturesetFeatures],
        population: ProtoTimedPopulation,
        result_file_format: ProtoFileFormat,
        options: ProtoOfflineServingQueryOptions,
        timeout_seconds: int = 60 * 60,
        poll_interval_seconds: int = 10,
    ) -> ProtoGetFeatureValuesResultResponse:
        request_handle: str = self.get_feature_values(
            features=features,
            population=population,
            result_file_format=result_file_format,
            options=options,
        )

        return self.poll_for_result(
            request_handle=request_handle,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
        )

    def get_feature_values_in_range_blocking(
        self,
        features: ProtoFeaturesetFeatures,
        lower_time_bound: ProtoTimestamp,
        upper_time_bound: ProtoTimestamp,
        result_file_format: ProtoFileFormat,
        options: ProtoOfflineServingQueryOptions,
        population: Optional[ProtoPopulation] = None,
        timeout_seconds: int = 60 * 60,
        poll_interval_seconds: int = 10,
    ) -> ProtoGetFeatureValuesResultResponse:
        request_handle: str = self.get_feature_values_in_range(
            features=features,
            lower_time_bound=lower_time_bound,
            upper_time_bound=upper_time_bound,
            population=population,
            result_file_format=result_file_format,
            options=options,
        )

        return self.poll_for_result(
            request_handle=request_handle,
            timeout_seconds=timeout_seconds,
            poll_interval_seconds=poll_interval_seconds,
        )



from abc import ABC, abstractmethod
from dataclasses import dataclass

from _qwak_proto.qwak.feature_store.sinks.sink_pb2 import (
    StreamingSink as ProtoStreamingSink,
)


@dataclass
class BaseSink(ABC):
    name: str

    @abstractmethod
    def to_proto_streaming_sink(self) -> ProtoStreamingSink:
        pass

    def __post_init__(self):
        pass



from qwak.inner.runtime_di import QwakRuntimeContainer, ContainerLock

from .run_model_locally import run_local


def wire_runtime_with_run_local():
    container = QwakRuntimeContainer()
    from qwak.model import decorators

    container.wire(
        packages=[
            decorators,
        ]
    )
    return container


# Since DI allows overriding containers that are already injected, and there isn't a built-in mechanism to prevent this
# we've added a lock class (singleton) that operates as a lock. Now the production override can lock the container, and
# it will prevent issues where the container is redefined with the basic non production container
if not ContainerLock.locked:
    wire_runtime_with_run_local()

__all__ = ["run_local", "wire_runtime_with_run_local"]



from collections import defaultdict
from dataclasses import dataclass, field
from typing import Dict, List, Tuple, Union

from _qwak_proto.qwak.batch_job.v1.batch_job_resources_pb2 import (
    NVIDIA_A10G,
    NVIDIA_A100,
    NVIDIA_K80,
    NVIDIA_T4,
    NVIDIA_V100,
)
from _qwak_proto.qwak.batch_job.v1.batch_job_service_pb2 import (
    CSV_INPUT_FILE_TYPE,
    CSV_OUTPUT_FILE_TYPE,
    FEATHER_INPUT_FILE_TYPE,
    FEATHER_OUTPUT_FILE_TYPE,
    PARQUET_INPUT_FILE_TYPE,
    PARQUET_OUTPUT_FILE_TYPE,
)
from qwak.exceptions import QwakException
from qwak.inner.tool.run_config import (
    ConfigCliMap,
    QwakConfigBase,
    YamlConfigMixin,
    validate_float,
    validate_int,
    validate_list_of_strings,
    validate_string,
)

DEFAULT_OUTPUT_FORMAT = "CSV"
OUTPUT_FORMATTERS_MAP = {
    DEFAULT_OUTPUT_FORMAT: CSV_OUTPUT_FILE_TYPE,
    "PARQUET": PARQUET_OUTPUT_FILE_TYPE,
    "FEATHER": FEATHER_OUTPUT_FILE_TYPE,
}

DEFAULT_INPUT_FORMAT = "CSV"
INPUT_FORMATTERS_MAP = {
    DEFAULT_INPUT_FORMAT: CSV_INPUT_FILE_TYPE,
    "PARQUET": PARQUET_INPUT_FILE_TYPE,
    "FEATHER": FEATHER_INPUT_FILE_TYPE,
}

GPU_TYPE_MAP = {
    "NVIDIA_V100": NVIDIA_V100,
    "NVIDIA_T4": NVIDIA_T4,
    "NVIDIA_K80": NVIDIA_K80,
    "NVIDIA_A100": NVIDIA_A100,
    "NVIDIA_A10G": NVIDIA_A10G,
}

PURCHASE_OPTION_SET = (
    "on-demand",
    "spot",
)


CONFIG_MAPPING: List[ConfigCliMap] = [
    ConfigCliMap("model_id", "execution.model_id", validate_string, True),
    ConfigCliMap("build_id", "execution.build_id", validate_string, False),
    ConfigCliMap("branch", "execution.branch", validate_string),
    ConfigCliMap("source_bucket", "execution.source_bucket", validate_string, False),
    ConfigCliMap("source_folder", "execution.source_folder", validate_string, False),
    ConfigCliMap("bucket", "execution.bucket", validate_string, False),
    ConfigCliMap(
        "destination_bucket", "execution.destination_bucket", validate_string, False
    ),
    ConfigCliMap(
        "destination_folder", "execution.destination_folder", validate_string, False
    ),
    ConfigCliMap(
        "input_file_type",
        "execution.input_file_type",
        validate_string,
        False,
    ),
    ConfigCliMap(
        "output_file_type",
        "execution.output_file_type",
        validate_string,
        False,
    ),
    ConfigCliMap("param_list", "execution.parameters", validate_list_of_strings, False),
    ConfigCliMap("job_timeout", "execution.job_timeout", validate_int, False),
    ConfigCliMap("file_timeout", "execution.file_timeout", validate_int, False),
    ConfigCliMap(
        "access_token_name", "execution.access_token_name", validate_string, False
    ),
    ConfigCliMap(
        "access_secret_name", "execution.access_secret_name", validate_string, False
    ),
    ConfigCliMap(
        "service_account_key_secret_name",
        "execution.service_account_key_secret_name",
        validate_string,
        False,
    ),
    ConfigCliMap("pods", "resources.pods", validate_int, False),
    ConfigCliMap("cpus", "resources.cpus", validate_float, False),
    ConfigCliMap("memory", "resources.memory", validate_int, False),
    ConfigCliMap("gpu_type", "resources.gpu_type", validate_string, False),
    ConfigCliMap("gpu_amount", "resources.gpu_amount", validate_int, False),
    ConfigCliMap("instance", "resources.instance_size", validate_string, False),
    ConfigCliMap(
        "purchase_option", "advanced_options.purchase_option", validate_string, False
    ),
]


@dataclass
class ExecutionConfig(YamlConfigMixin, QwakConfigBase):
    def merge_bucket_arguments(self):
        if self.execution.bucket:
            if not self.execution.source_bucket:
                self.execution.source_bucket = self.execution.bucket
            if not self.execution.destination_bucket:
                self.execution.destination_bucket = self.execution.bucket

    def __post_init__(self):
        self.merge_bucket_arguments()

    def _post_merge_cli(self):
        self.merge_bucket_arguments()
        self.execution.parameters = dictify_params(self.execution.parameters)

        if not self.execution.source_bucket:
            raise QwakException(
                "Must supply either --bucket or --source-bucket parameter"
            )
        if not self.execution.destination_bucket:
            raise QwakException(
                "Must supply either --bucket or --destination-bucket parameter"
            )

    @property
    def _config_mapping(self) -> List[ConfigCliMap]:
        return CONFIG_MAPPING

    @dataclass
    class Execution:
        model_id: str = field(default="")
        build_id: str = field(default="")
        branch: str = field(default="main")
        bucket: str = field(default="")
        source_bucket: str = field(default="")
        source_folder: str = field(default="")
        destination_bucket: str = field(default="")
        destination_folder: str = field(default="")
        input_file_type: str = field(default=DEFAULT_INPUT_FORMAT)
        output_file_type: str = field(default=DEFAULT_OUTPUT_FORMAT)
        access_token_name: str = field(default="")
        access_secret_name: str = field(default="")
        job_timeout: int = field(default=0)
        file_timeout: int = field(default=0)
        parameters: dict = field(default_factory=lambda: defaultdict(str))
        service_account_key_secret_name: str = field(default="")

    @dataclass
    class Warmup:
        model_id: str = field(default="")
        branch: str = field(default="main")
        timeout: int = field(default=60 * 60)

    @dataclass
    class Resources:
        pods: int = field(default=0)
        cpus: float = field(default=0)
        memory: int = field(default=0)
        gpu_amount: int = field(default=0)
        gpu_type: str = field(default="")
        instance_size: str = field(default="")

    @dataclass
    class AdvancedOptions:
        custom_iam_role_arn: str = field(default=None)
        purchase_option: str = field(default=None)
        service_account_key_secret_name: str = field(default=None)

    warmup: Warmup = field(default_factory=Warmup)
    execution: Execution = field(default_factory=Execution)
    resources: Resources = field(default_factory=Resources)
    advanced_options: AdvancedOptions = field(default_factory=AdvancedOptions)


def dictify_params(parameters: Union[Dict[str, str], List[str], Tuple[str]]):
    if isinstance(parameters, dict):
        return parameters

    result = dict()

    if isinstance(parameters, (list, tuple)):
        for param in parameters:
            if "=" not in param:
                raise QwakException(
                    f'The parameter definition passed {param} is invalid. Format is "KEY=VALUE"'
                )
            split_param = param.split("=")
            result[split_param[0]] = split_param[1]

    return result
